==> dpsearch-4.28.tar.gz: dpsearch-4.28/src/boolean.c <==
/* Copyright (C) 2003-2004 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "dps_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "dps_common.h"
#include "dps_boolean.h"
#include "dps_utils.h"

/*
#define DEBUG_BOOL
*/

#ifdef DEBUG_BOOL
static char item_type(int cmd, unsigned long arg){
	switch(cmd){
		case DPS_STACK_LEFT:	return('(');
		case DPS_STACK_RIGHT:	return(')');
		case DPS_STACK_OR:	return('|');
		case DPS_STACK_AND:	return('&');
		case DPS_STACK_NOT:	return('~');
		case DPS_STACK_PHRASE:	return('"');
		case DPS_STACK_WORD:	return(arg?'1':'0');
		default:		return('?');
	}
}
#endif

static int TOPCMD(DPS_BOOLSTACK * s){
	if(s->ncstack)
		return(s->cstack[s->ncstack-1]);
	else
		return(DPS_STACK_BOT);
}

static int POPARG(DPS_BOOLSTACK * s){
	if(s->nastack>0){
		s->nastack--;
		return(s->astack[s->nastack]);
	}else{
		return(DPS_STACK_BOT);
	}
}

static int POPCMD(DPS_BOOLSTACK * s){
	if(s->ncstack>0){
		s->ncstack--;
		return(s->cstack[s->ncstack]);
	}else{
		return(DPS_STACK_BOT);
	}
}

static int PUSHARG(DPS_BOOLSTACK * s,unsigned long arg){
	s->astack[s->nastack]=arg;
	s->nastack++;
	if (s->nastack >= s->mastack) {
	  s->mastack += DPS_MAXSTACK;
	  s->astack = (unsigned long*)DpsRealloc(s->astack, s->mastack * sizeof(unsigned long));
	  if (s->astack == NULL) return DPS_ERROR;
	}
	return DPS_OK;
}

static int PUSHCMD(DPS_BOOLSTACK * s,int arg){
	s->cstack[s->ncstack]=arg;
	s->ncstack++;
	if (s->ncstack >= s->mcstack) {
	  s->mcstack += DPS_MAXSTACK;
	  s->cstack = (int*)DpsRealloc(s->cstack, s->mcstack * sizeof(int));
	  if (s->cstack == NULL) return DPS_ERROR;
	}
	return DPS_OK;
}

/* Compute one operation and store result */
static int perform(DPS_BOOLSTACK * s,int com){
	unsigned long res,x1,x2;
	int rc = DPS_OK;
	switch(com){
		case DPS_STACK_OR:
			/* This is to avoid compiler      */
			/* optimization in || calculation */
			x1=POPARG(s);
			x2=POPARG(s);
			res=(x1||x2);
#ifdef DEBUG_BOOL
			fprintf(stderr,"Perform %lu | %lu\n",x1,x2);
#endif
			rc = PUSHARG(s, res);
			break;
		case DPS_STACK_AND:
			x1=POPARG(s);
			x2=POPARG(s);
			res=(x1&&x2);
#ifdef DEBUG_BOOL
			fprintf(stderr,"Perform %lu & %lu\n",x1,x2);
#endif
			rc = PUSHARG(s, res);
			break;
		case DPS_STACK_NOT:
			x1=POPARG(s);
			res=x1?0:1;
#ifdef DEBUG_BOOL
			fprintf(stderr,"Perform ~%lu\n",x1);
#endif
			rc = PUSHARG(s, res);
			break;
	}
	return rc;
}


/* Main function to calculate items sequence */
int DpsCalcBoolItems(DPS_STACK_ITEM * items, size_t nitems, size_t *count){
	DPS_BOOLSTACK s;
	int res;
	size_t i;

	/* Init stack */
	s.ncstack=0;
	s.nastack=0;
	s.mcstack = s.mastack = DPS_MAXSTACK;
	s.cstack = (int*)DpsMalloc(DPS_MAXSTACK * sizeof(int));
	if (s.cstack == NULL) return DPS_STACK_ERR;
	s.astack = (unsigned long*)DpsMalloc(DPS_MAXSTACK * sizeof(unsigned long));
	if (s.astack == NULL) {
	  DPS_FREE(s.cstack);
	  return DPS_STACK_ERR;
	}

#ifdef DEBUG_BOOL
	fprintf(stderr,"--------\n");
#endif
	for(i=0;i<nitems;i++){
		int c;
#ifdef DEBUG_BOOL
		fprintf(stderr,".[%d] %c\n", i, item_type(items[i].cmd, count[items[i].arg]));
#endif
		switch(c=items[i].cmd){
			case DPS_STACK_RIGHT:
				/* Perform till LEFT bracket */
				while((TOPCMD(&s)!=DPS_STACK_LEFT)&&(TOPCMD(&s)!=DPS_STACK_BOT))
				  if (DPS_OK != perform(&s, POPCMD(&s))) {
				    DPS_FREE(s.cstack);
				    DPS_FREE(s.astack);
				    return DPS_STACK_ERR;
				  }
				/* Pop LEFT bracket itself */
				if(TOPCMD(&s)==DPS_STACK_LEFT)
					POPCMD(&s);
				break;
			case DPS_STACK_OR:
			case DPS_STACK_AND:
				while(c<=TOPCMD(&s)){
					if (DPS_OK != perform(&s,POPCMD(&s))) {
					  DPS_FREE(s.cstack);
					  DPS_FREE(s.astack);
					  return DPS_STACK_ERR;
					}
				}
				/* IMPORTANT! No break here! That's OK*/
				/*   ! */
			case DPS_STACK_LEFT:
			case DPS_STACK_NOT:
			        if (PUSHCMD(&s,c) != DPS_OK) {
				  DPS_FREE(s.cstack);
				  DPS_FREE(s.astack);
				  return DPS_STACK_ERR;
				}
				break;
  		        case DPS_STACK_PHRASE:
			        if (DPS_OK != PUSHARG(&s, ( (i + 1 < nitems) && count[items[i+1].arg] ) ? 1UL : 0UL)) {
				  DPS_FREE(s.cstack);
				  DPS_FREE(s.astack);
				  return DPS_STACK_ERR;
				}
			        for (i++; items[i].cmd != DPS_STACK_PHRASE; i++)
#ifdef DEBUG_BOOL
		fprintf(stderr,"[%d] %c\n", i, item_type(items[i].cmd, count[items[i].arg]));
		fprintf(stderr,"[%d] %c\n", i, item_type(items[i].cmd, count[items[i].arg]));
#endif
 		                ;
				break;
			case DPS_STACK_WORD:
			default:
			        if (DPS_OK != PUSHARG(&s, (count[items[i].arg]) ? 1UL : 0UL)) {
				  DPS_FREE(s.cstack);
				  DPS_FREE(s.astack);
				  return DPS_STACK_ERR;
				}
				break;
		}
	}
	while(TOPCMD(&s)!=DPS_STACK_BOT){
	  if (DPS_OK != perform(&s, POPCMD(&s))) {
	    DPS_FREE(s.cstack);
	    DPS_FREE(s.astack);
	    return DPS_STACK_ERR;
	  }
	}
	res=POPARG(&s);
	DPS_FREE(s.cstack); DPS_FREE(s.astack);
#ifdef DEBUG_BOOL
	fprintf(stderr, "result: %d\n", res);
#endif
	return(res);
}

==> dpsearch-4.29.tar.gz: dpsearch-4.29/src/boolean.c <==
/* Copyright (C) 2003-2004 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/
#include "dps_common.h"
#include "dps_boolean.h"
#include "dps_utils.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

/*
#define DEBUG_BOOL
*/

#ifdef DEBUG_BOOL
static char item_type(int cmd, unsigned long arg){
	switch(cmd){
		case DPS_STACK_LEFT:	return('(');
		case DPS_STACK_RIGHT:	return(')');
		case DPS_STACK_OR:	return('|');
		case DPS_STACK_AND:	return('&');
		case DPS_STACK_NOT:	return('~');
		case DPS_STACK_PHRASE:	return('"');
		case DPS_STACK_WORD:	return(arg?'1':'0');
		default:		return('?');
	}
}
#endif


DPS_BOOLSTACK *DpsBoolStackInit(DPS_BOOLSTACK *s) {
	if(s == NULL) {
		s = (DPS_BOOLSTACK*)DpsMalloc(sizeof(DPS_BOOLSTACK));
		if (s == NULL) return NULL;
		bzero((void*)s, sizeof(*s));
		s->freeme = 1;
	}else{
		bzero((void*)s, sizeof(*s));
	}
	s->ncstack = 0;
	s->nastack = 0;
	s->mcstack = s->mastack = DPS_MAXSTACK;
	s->cstack = (int*)DpsMalloc(DPS_MAXSTACK * sizeof(int));
	if (s->cstack == NULL) { if (s->freeme) DPS_FREE(s); return NULL; }
	s->astack = (unsigned long*)DpsMalloc(DPS_MAXSTACK * sizeof(unsigned long));
	if (s->astack == NULL) {
	  DPS_FREE(s->cstack);
	  if (s->freeme) DPS_FREE(s); 
	  return NULL;
	}
	return s;
}


void DpsBoolStackFree(DPS_BOOLSTACK *s) {
	DPS_FREE(s->cstack); DPS_FREE(s->astack);
	if (s->freeme) DPS_FREE(s); 
}


static int TOPCMD(DPS_BOOLSTACK * s){
	if(s->ncstack)
		return(s->cstack[s->ncstack-1]);
	else
		return(DPS_STACK_BOT);
}

static int POPARG(DPS_BOOLSTACK * s){
	if(s->nastack>0){
		s->nastack--;
		return(s->astack[s->nastack]);
	}else{
		return(DPS_STACK_BOT);
	}
}

static int POPCMD(DPS_BOOLSTACK * s){
	if(s->ncstack>0){
		s->ncstack--;
		return(s->cstack[s->ncstack]);
	}else{
		return(DPS_STACK_BOT);
	}
}

static int PUSHARG(DPS_BOOLSTACK * s,unsigned long arg){
	s->astack[s->nastack]=arg;
	s->nastack++;
	if (s->nastack >= s->mastack) {
	  s->mastack += DPS_MAXSTACK;
	  s->astack = (unsigned long*)DpsRealloc(s->astack, s->mastack * sizeof(unsigned long));
	  if (s->astack == NULL) return DPS_ERROR;
	}
	return DPS_OK;
}

static int PUSHCMD(DPS_BOOLSTACK * s,int arg){
	s->cstack[s->ncstack]=arg;
	s->ncstack++;
	if (s->ncstack >= s->mcstack) {
	  s->mcstack += DPS_MAXSTACK;
	  s->cstack = (int*)DpsRealloc(s->cstack, s->mcstack * sizeof(int));
	  if (s->cstack == NULL) return DPS_ERROR;
	}
	return DPS_OK;
}

/* Compute one operation and store result */
static int perform(DPS_BOOLSTACK * s,int com){
	unsigned long res,x1,x2;
	int rc = DPS_OK;
	switch(com){
		case DPS_STACK_OR:
			/* This is to avoid compiler      */
			/* optimization in || calculation */
			x1=POPARG(s);
			x2=POPARG(s);
			res=(x1||x2);
#ifdef DEBUG_BOOL
			fprintf(stderr,"Perform %lu | %lu\n",x1,x2);
#endif
			rc = PUSHARG(s, res);
			break;
		case DPS_STACK_AND:
			x1=POPARG(s);
			x2=POPARG(s);
			res=(x1&&x2);
#ifdef DEBUG_BOOL
			fprintf(stderr,"Perform %lu & %lu\n",x1,x2);
#endif
			rc = PUSHARG(s, res);
			break;
		case DPS_STACK_NOT:
			x1=POPARG(s);
			res=x1?0:1;
#ifdef DEBUG_BOOL
			fprintf(stderr,"Perform ~%lu\n",x1);
#endif
			rc = PUSHARG(s, res);
			break;
	}
	return rc;
}


/* Main function to calculate items sequence */
int DpsCalcBoolItems(DPS_BOOLSTACK *s, DPS_STACK_ITEM * items, size_t nitems, size_t *count){
	int res;
	size_t i;

	/* Init stack */
	s->ncstack = 0;
	s->nastack = 0;

#ifdef DEBUG_BOOL
	fprintf(stderr,"--------\n");
#endif
	for(i=0;i<nitems;i++){
		int c;
#ifdef DEBUG_BOOL
		fprintf(stderr,".[%d] %c\n", i, item_type(items[i].cmd, count[items[i].arg]));
#endif
		switch(c=items[i].cmd){
			case DPS_STACK_RIGHT:
				/* Perform till LEFT bracket */
				while((TOPCMD(s) != DPS_STACK_LEFT) && (TOPCMD(s) != DPS_STACK_BOT))
				  if (DPS_OK != perform(s, POPCMD(s))) {
				    return DPS_STACK_ERR;
				  }
				/* Pop LEFT bracket itself */
				if(TOPCMD(s) == DPS_STACK_LEFT)
					POPCMD(s);
				break;
			case DPS_STACK_OR:
			case DPS_STACK_AND:
				while(c <= TOPCMD(s)) {
					if (DPS_OK != perform(s, POPCMD(s))) {
					  return DPS_STACK_ERR;
					}
				}
				/* IMPORTANT! No break here! That's OK*/
				/*   ! */
			case DPS_STACK_LEFT:
			case DPS_STACK_NOT:
			        if (PUSHCMD(s,c) != DPS_OK) {
				  return DPS_STACK_ERR;
				}
				break;
  		        case DPS_STACK_PHRASE:
			        if (DPS_OK != PUSHARG(s, ( (i + 1 < nitems) && count[items[i+1].arg] ) ? 1UL : 0UL)) {
				  return DPS_STACK_ERR;
				}
			        for (i++; items[i].cmd != DPS_STACK_PHRASE; i++)
#ifdef DEBUG_BOOL
		fprintf(stderr,"[%d] %c\n", i, item_type(items[i].cmd, count[items[i].arg]));
		fprintf(stderr,"[%d] %c\n", i, item_type(items[i].cmd, count[items[i].arg]));
#endif
 		                ;
				break;
			case DPS_STACK_WORD:
			default:
			        if (DPS_OK != PUSHARG(s, (count[items[i].arg]) ? 1UL : 0UL)) {
				  return DPS_STACK_ERR;
				}
				break;
		}
	}
	while(TOPCMD(s) != DPS_STACK_BOT) {
	  if (DPS_OK != perform(s, POPCMD(s))) {
	    return DPS_STACK_ERR;
	  }
	}
	res = POPARG(s);
#ifdef DEBUG_BOOL
	fprintf(stderr, "result: %d\n", res);
#endif
	return(res);
}
